import pytest
from ngsolve.meshes import *
from netgen.csg import *
from ngsolve import *
from xfem import *
from math import pi
import netgen.meshing as ngm
from netgen.geom2d import SplineGeometry
from xfem.lset_spacetime import *

tref = ReferenceTimeVariable()

@pytest.mark.parametrize("quad", [True, False])
@pytest.mark.parametrize("integrands", [(tref,0.5,0,1),
                                        (tref**3,0.25,0,3),
                                        ((1-tref)**3,0.25,0,3),
                                        (x,0.5,1,0),
                                        (tref*tref*(x*x+y*y),2/9,2,2)])
def test_spacetime_integrate_no_cut(quad, integrands):
    mesh = MakeStructured2DMesh(quads = quad, nx=1, ny=1)    

    f,ref_value, space_order, time_order = integrands
    
    h1fes = H1(mesh,order=1)
    tfe = ScalarTimeFE(1) 
    fes= SpaceTimeFESpace(h1fes,tfe)
    lset_approx = GridFunction(fes)

    lset_approx.vec[:] = -1

    integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : NEG},
                         cf=f, mesh=mesh, order = space_order, time_order=time_order)
    print("Integral: ", integral)
    error = abs(integral - ref_value)
    
    assert error < 5e-15


@pytest.mark.parametrize("quad", [True, False])
@pytest.mark.parametrize("domain", [NEG, POS, IF])
def test_spacetime_integrateX_via_straight_cutted_quad2Dplus1D(domain, quad):
    mesh = MakeStructured2DMesh(quads = quad, nx=1, ny=1)    

    tref = ReferenceTimeVariable()
    
    levelset = lambda t : 1 - 2*x - 2*t
    referencevals = { POS : 1./8, NEG : 1 - 1/8, IF : 1.0/2 }

    h1fes = H1(mesh,order=1)
    lset_approx_h1 = GridFunction(h1fes)
    tfe = ScalarTimeFE(1) 
    fes= SpaceTimeFESpace(h1fes,tfe)
    lset_approx = GridFunction(fes)

    InterpolateToP1(levelset(0),lset_approx_h1)
    lset_approx.vec[0:h1fes.ndof].data = lset_approx_h1.vec
    InterpolateToP1(levelset(1),lset_approx_h1)
    lset_approx.vec[h1fes.ndof:2*h1fes.ndof].data = lset_approx_h1.vec

    print(lset_approx.vec)
    
    f = CoefficientFunction(1)
    
    integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain},
                         cf=f, mesh=mesh, order = 0, time_order=0)
    print("Integral: ", integral)
    error = abs(integral - referencevals[domain])
    
    assert error < 5e-15

@pytest.mark.parametrize("pitfal1", [False])
@pytest.mark.parametrize("pitfal2", [False])
@pytest.mark.parametrize("pitfal3", [False])

def test_spacetime_model_spacetime(pitfal1, pitfal2, pitfal3):
    square = SplineGeometry()
    square.AddRectangle([0,0],[1,1],bc=1)
    ngmesh = square.GenerateMesh(maxh=0.05, quad_dominated=False)
    mesh = Mesh (ngmesh)
    
    fes1 = V=H1(mesh, order=1, dirichlet=[1,2,3,4])
    k_t = 1
    tfe = ScalarTimeFE(k_t) 
    
    st_fes = SpaceTimeFESpace(fes1,tfe)
    st_fes_ic = SpaceTimeFESpace(fes1,tfe)
    
    tend = 1.0
    delta_t = 1/32
    
    told = Parameter(0)
    tref = ReferenceTimeVariable()
    t = told + delta_t*tref

    u_exact = lambda t: CoefficientFunction( sin(pi*t)*sin(pi*x)*sin(pi*x)*sin(pi*y)*sin(pi*y)  )
    coeff_f = CoefficientFunction( pi*cos(pi*t)*sin(pi*x)*sin(pi*x)*sin(pi*y)*sin(pi*y)
                               -2*pi*pi*sin(pi*t)*( cos(pi*x)*cos(pi*x)*sin(pi*y)*sin(pi*y)              
                                                   -sin(pi*x)*sin(pi*x)*sin(pi*y)*sin(pi*y)
                                                   +cos(pi*y)*cos(pi*y)*sin(pi*x)*sin(pi*x)
                                                  -sin(pi*x)*sin(pi*x)*sin(pi*y)*sin(pi*y))) 

    u0 = GridFunction(st_fes)
    u0_ic = GridFunction(fes1)
    u = st_fes.TrialFunction()
    v = st_fes.TestFunction()

    # dummy lset domain to call symboliccutbfi instead of usual symbolicbfi...
    levelset = (sqrt(x*x+y*y) - 1000.5)
    lsetp1 = GridFunction(H1(mesh,order=1))
    InterpolateToP1(levelset,lsetp1)
    lset_neg = { "levelset" : lsetp1, "domain_type" : NEG, "subdivlvl" : 0}

    a = BilinearForm(st_fes,symmetric=False)
    a += SymbolicBFI(levelset_domain = lset_neg, form = delta_t*grad(u)*grad(v), time_order=2)
    a += SymbolicBFI(form = fix_tref(u,0)*fix_tref(v,0) )
    a += SymbolicBFI(levelset_domain = lset_neg, form = dt(u)*v, time_order=2)
    a.Assemble()

    t_old = 0
    u0_ic.Set(u_exact(0))
    if pitfal1:
        u0_ic.Set(u_exact(t))
    
    while tend - t_old > delta_t/2:
        f = LinearForm(st_fes)
        f += SymbolicLFI(levelset_domain = lset_neg, form = delta_t*coeff_f*v, time_order=2)
        f += SymbolicLFI(form = u0_ic*fix_tref(v,0))
        if pitfal2:
            f += SymbolicLFI(form = u0_ic*v)
        f.Assemble()
        
        u0.vec.data = a.mat.Inverse(st_fes.FreeDofs(),"") * f.vec
        
        # exploiting the nodal property of the time fe:
        #u0_ic.vec[:] = u0.vec[0:fes1.ndof]
        u0_ic.vec[:].data = u0.vec[fes1.ndof : 2*fes1.ndof]
    
        t_old = t_old + delta_t
        told.Set(t_old)
        
        l2error = sqrt (Integrate ( (u_exact(t_old) -u0_ic)**2, mesh))
        if pitfal3:
            l2error = sqrt (Integrate ( (u_exact(t) -u0_ic)**2, mesh))
                
        print("t = {0}, l2error = {1}".format(t_old,l2error))
        assert l2error < 5e-3
    assert l2error < 2e-4

def test_spacetime_model_spacetime_caller():
    try:
        test_spacetime_model_spacetime(True, False, False)
    except Exception as e:
        if("TimeVariableCoefficientFunction::Evaluate called with a mere space IR" in str(e)):
            print("Failed properly")
        else:
            print('Unexpected exception raised:', e)
            raise Exception("Wrong kind of failure")
    else:
        raise Exception("No failure at all")
    
    try:
        test_spacetime_model_spacetime(False, True, False)
    except Exception as e:
        if("SpaceTimeFE :: CalcShape called with a mere space IR" in str(e)):
            print("Failed properly")
        else:
            print('Unexpected exception raised:', e)
            raise Exception("Wrong kind of failure")
    else:
        raise Exception("No failure at all")
    
    try:
        test_spacetime_model_spacetime(False, False, True)
    except Exception as e:
        if("TimeVariableCoefficientFunction::Evaluate called with a mere space IR" in str(e)):
            print("Failed properly")
        else:
            print('Unexpected exception raised:', e)
            raise Exception("Wrong kind of failure")
    else:
        raise Exception("No failure at all")

# def test_spacetime_spaceP1_timeCGP1():
#     ngsglobals.msg_level = 1

#     square = SplineGeometry()
#     square.AddRectangle([-1,-1],[1,1])
#     ngmesh = square.GenerateMesh(maxh=0.08, quad_dominated=False)
#     mesh = Mesh (ngmesh)

#     coef_told = Parameter(0)
#     coef_delta_t = Parameter(0)
#     tref = ReferenceTimeVariable()
#     t = coef_told + coef_delta_t*tref

#     r0 = 0.5

#     # position shift of the geometry in time
#     rho =  CoefficientFunction((1/(pi))*sin(2*pi*t))
#     rhoL = lambda t:CoefficientFunction((1/(pi))*sin(2*pi*t))
#     #convection velocity:
#     d_rho = CoefficientFunction(2*cos(2*pi*t))
#     w = CoefficientFunction((0,d_rho)) 

#     # level set
#     r = sqrt(x**2+(y-rho)**2)
#     levelset= r - r0

#     # diffusion coefficient
#     alpha = 1

#     # solution and r.h.s.
#     Q = pi/r0   
#     u_exact = cos(Q*r) * sin(pi*t)
#     u_exactL = lambda t: cos(Q*sqrt(x**2+(y-rhoL(t))**2)) * sin(pi*t)
#     coeff_f = (Q/r * sin(Q*r) + (Q**2) * cos(Q*r)) * sin(pi*t) + pi * cos(Q*r) * cos(pi*t)
#     u_init = u_exact

#     # polynomial order in time
#     k_t = 1
#     # polynomial order in space
#     k_s = 1
#     # spatial FESpace for solution
#     fes1 = H1(mesh, order=k_s)
#     # polynomial order in time for level set approximation
#     lset_order_time = k_t
#     # integration order in time
#     time_order = 2*k_t
#     # time finite element (nodal!)
#     tfe = ScalarTimeFE(k_t) 
#     tfe_i = ScalarTimeFE(k_t, skip_first_node=True) # interior
#     tfe_e = ScalarTimeFE(k_t, only_first_node=True) # exterior (inital values)
#     tfe_t = ScalarTimeFE(k_t-1)                     # test

#     # space-time finite element space
#     st_fes = SpaceTimeFESpace(fes1,tfe, flags = {"dgjumps": True})
#     st_fes_i = SpaceTimeFESpace(fes1,tfe_i, flags = {"dgjumps": True})
#     st_fes_e = SpaceTimeFESpace(fes1,tfe_e, flags = {"dgjumps": True})
#     st_fes_t = SpaceTimeFESpace(fes1,tfe_t, flags = {"dgjumps": True})

#     #Fitted heat equation example
#     tend = 1
#     delta_t = tend/64
#     coef_delta_t.Set(delta_t)
#     tnew = 0
#     told = 0

#     lset_p1 = GridFunction(st_fes)

#     SpaceTimeInterpolateToP1(levelset,tref,lset_p1)

#     lset_top = CreateTimeRestrictedGF(lset_p1,1.0)
#     lset_bottom = CreateTimeRestrictedGF(lset_p1,0.0)

#     gfu_i = GridFunction(st_fes_i)
#     gfu_e = GridFunction(st_fes_e)

#     u_last = CreateTimeRestrictedGF(gfu_e,0)
#     SpaceTimeWeakSet(gfu_e, u_exactL(0.0), fes1)

#     u_i = st_fes_i.TrialFunction()
#     u_e = st_fes_e.TrialFunction()
#     v_t = st_fes_t.TestFunction()

#     h = specialcf.mesh_size

#     lset_neg = { "levelset" : lset_p1, "domain_type" : NEG, "subdivlvl" : 0}
#     lset_neg_bottom = { "levelset" : lset_bottom, "domain_type" : NEG, "subdivlvl" : 0}
#     lset_neg_top = { "levelset" : lset_top, "domain_type" : NEG, "subdivlvl" : 0}

#     def SpaceTimeNegBFI(form):
#         return SymbolicBFI(levelset_domain = lset_neg, form = form, time_order=time_order)

#     ci = CutInfo(mesh,time_order=time_order)

        
#     hasneg_integrators_a_i = []
#     hasneg_integrators_a_e = []
#     hasneg_integrators_f = []
#     patch_integrators_a_i = []
#     patch_integrators_a_e = []

#     for hasneg_integrators_a,u in [(hasneg_integrators_a_i,u_i),(hasneg_integrators_a_e,u_e)]:
#         hasneg_integrators_a.append(SpaceTimeNegBFI(form = -dt(v_t)*u))
#         hasneg_integrators_a.append(SpaceTimeNegBFI(form = -delta_t*InnerProduct(w,grad(v_t))*u))
#         hasneg_integrators_a.append(SpaceTimeNegBFI(form = delta_t*alpha*grad(u)*grad(v_t)))

#     for patch_integrators_a,u in [(patch_integrators_a_i,u_i),(patch_integrators_a_e,u_e)]:
#         patch_integrators_a.append(SymbolicFacetPatchBFI(form = delta_t*1.05*h**(-2)*(u-u.Other())*(v_t-v_t.Other()),
#                                                         skeleton=False, time_order=time_order))

#     hasneg_integrators_a_i.append(SymbolicBFI(levelset_domain = lset_neg_top, form = fix_tref(u_i,1)*fix_tref(v_t,1)))

#     hasneg_integrators_a_e.append(SymbolicBFI(levelset_domain = lset_neg_bottom, form = -fix_tref(u_e,0)*fix_tref(v_t,0)))
#     #hasneg_integrators_f.append(SymbolicLFI(levelset_domain = lset_neg_bottom,form = u_last*fix_tref(v,0)))

#     hasneg_integrators_f.append(SymbolicLFI(levelset_domain = lset_neg, form = delta_t*coeff_f*v_t, time_order=time_order)) 

#     a_i = BilinearForm(trialspace = st_fes_i, testspace = st_fes_t, check_unused=False, symmetric=False)
#     for integrator in hasneg_integrators_a_i + patch_integrators_a_i:
#         a_i += integrator

#     a_e = BilinearForm(trialspace = st_fes_e, testspace = st_fes_t, check_unused=False, symmetric=False)
#     for integrator in hasneg_integrators_a_e + patch_integrators_a_e:
#         a_e += integrator

#     f = LinearForm(st_fes_t)

#     for integrator in hasneg_integrators_f:
#         f += integrator

#     while tend - told > delta_t/2:
#         SpaceTimeInterpolateToP1(levelset,tref,lset_p1)
#         RestrictGFInTime(spacetime_gf=lset_p1,reference_time=0.0,space_gf=lset_bottom)
#         RestrictGFInTime(spacetime_gf=lset_p1,reference_time=1.0,space_gf=lset_top)

#         # update markers in (space-time) mesh
#         ci.Update(lset_p1,time_order=time_order)

#         # re-compute the facets for stabilization:
#         ba_facets = GetFacetsWithNeighborTypes(mesh,a=ci.GetElementsOfType(HASNEG),
#                                                     b=ci.GetElementsOfType(IF))
#         # re-evaluate the "active dofs" in the space time slab
#         active_dofs = GetDofsOfElements(st_fes,ci.GetElementsOfType(HASNEG))

#         # re-set definedonelements-markers according to new markings:
#         for integrator in hasneg_integrators_a_i + hasneg_integrators_a_e + hasneg_integrators_f:
#             integrator.SetDefinedOnElements(ci.GetElementsOfType(HASNEG))
#         for integrator in patch_integrators_a:
#             integrator.SetDefinedOnElements(ba_facets)

#         # assemble linear system
#         a_i.Assemble()
#         a_e.Assemble()
#         f.Assemble()

#         # solve linear system
#         inv = a_i.mat.Inverse(active_dofs,inverse="")
#         f.vec.data -= a_e.mat * gfu_e.vec
#         gfu_i.vec.data =  inv * f.vec

#         # evaluate upper trace of solution for
#         #  * for error evaluation 
#         #  * upwind-coupling to next time slab
#         RestrictGFInTime(spacetime_gf=gfu_i,reference_time=1.0,space_gf=u_last)   
        
#         SpaceTimeWeakSet(gfu_e, u_last, fes1)
        
#         # update time variable (float and ParameterCL)
#         told = told + delta_t
#         coef_told.Set(told)
        
#         # compute error at end of time slab
#         l2error = sqrt(Integrate(lset_neg_top,(u_exactL(told) -u_last)**2,mesh))
#         # print time and error
#         print("t = {0:10}, l2error = {1:20}".format(told,l2error),end="\n")
#         assert(l2error < 0.3)
#     assert(l2error < 0.08)

def test_spacetime_spaceP1_timeDGP1():
    ngsglobals.msg_level = 1

    square = SplineGeometry()
    square.AddRectangle([-1,-1],[1,1])
    ngmesh = square.GenerateMesh(maxh=0.08, quad_dominated=False)
    mesh = Mesh (ngmesh)

    coef_told = Parameter(0)
    coef_delta_t = Parameter(0)
    tref = ReferenceTimeVariable()
    t = coef_told + coef_delta_t*tref

    r0 = 0.5

    # position shift of the geometry in time
    rho =  CoefficientFunction((1/(pi))*sin(2*pi*t))
    rhoL = lambda t:CoefficientFunction((1/(pi))*sin(2*pi*t))
    #convection velocity:
    d_rho = CoefficientFunction(2*cos(2*pi*t))
    w = CoefficientFunction((0,d_rho)) 

    # level set
    r = sqrt(x**2+(y-rho)**2)
    levelset= r - r0

    # diffusion coefficient
    alpha = 1

    # solution and r.h.s.
    Q = pi/r0   
    u_exact = cos(Q*r) * sin(pi*t)
    u_exactL = lambda t: cos(Q*sqrt(x**2+(y-rhoL(t))**2)) * sin(pi*t)
    coeff_f = (Q/r * sin(Q*r) + (Q**2) * cos(Q*r)) * sin(pi*t) + pi * cos(Q*r) * cos(pi*t)
    u_init = u_exact

    # polynomial order in time
    k_t = 1
    # polynomial order in space
    k_s = 1
    # spatial FESpace for solution
    fes1 = H1(mesh, order=k_s)
    # polynomial order in time for level set approximation
    lset_order_time = 1
    # integration order in time
    time_order = 2
    # time finite element (nodal!)
    tfe = ScalarTimeFE(k_t) 
    # space-time finite element space
    st_fes = SpaceTimeFESpace(fes1,tfe, flags = {"dgjumps": True})

    #Unfitted heat equation example
    tend = 1
    delta_t = tend/32
    coef_delta_t.Set(delta_t)
    tnew = 0
    told = 0

    lset_p1 = GridFunction(st_fes)

    SpaceTimeInterpolateToP1(levelset,tref,lset_p1)

    lset_top = CreateTimeRestrictedGF(lset_p1,1.0)
    lset_bottom = CreateTimeRestrictedGF(lset_p1,0.0)

    gfu = GridFunction(st_fes)

    u_last = CreateTimeRestrictedGF(gfu,0)
    u_last.Set(u_exactL(0.))

    u,v = st_fes.TnT()

    h = specialcf.mesh_size

    lset_neg = { "levelset" : lset_p1, "domain_type" : NEG, "subdivlvl" : 0}
    lset_neg_bottom = { "levelset" : lset_bottom, "domain_type" : NEG, "subdivlvl" : 0}
    lset_neg_top = { "levelset" : lset_top, "domain_type" : NEG, "subdivlvl" : 0}

    def SpaceTimeNegBFI(form):
        return SymbolicBFI(levelset_domain = lset_neg, form = form, time_order=time_order)

    ci = CutInfo(mesh,time_order=time_order)

        
    hasneg_integrators_a = []
    hasneg_integrators_f = []
    patch_integrators_a = []
    hasneg_integrators_a.append(SpaceTimeNegBFI(form = delta_t*alpha*grad(u)*grad(v)))
    hasneg_integrators_a.append(SymbolicBFI(levelset_domain = lset_neg_top, form = fix_tref(u,1)*fix_tref(v,1)))
    hasneg_integrators_a.append(SpaceTimeNegBFI(form = -u*dt(v)))
    hasneg_integrators_a.append(SpaceTimeNegBFI(form = -delta_t*u*InnerProduct(w,grad(v))))
    patch_integrators_a.append(SymbolicFacetPatchBFI(form = delta_t*1.05*h**(-2)*(u-u.Other())*(v-v.Other()),
                                                    skeleton=False, time_order=time_order))
    hasneg_integrators_f.append(SymbolicLFI(levelset_domain = lset_neg, form = delta_t*coeff_f*v, time_order=time_order)) 
    hasneg_integrators_f.append(SymbolicLFI(levelset_domain = lset_neg_bottom,form = u_last*fix_tref(v,0)))


    a = BilinearForm(st_fes,check_unused=False,symmetric=False)
    for integrator in hasneg_integrators_a + patch_integrators_a:
        a += integrator

    f = LinearForm(st_fes)

    for integrator in hasneg_integrators_f:
        f += integrator

    while tend - told > delta_t/2:
        SpaceTimeInterpolateToP1(levelset,tref,lset_p1)
        RestrictGFInTime(spacetime_gf=lset_p1,reference_time=0.0,space_gf=lset_bottom)
        RestrictGFInTime(spacetime_gf=lset_p1,reference_time=1.0,space_gf=lset_top)

        # update markers in (space-time) mesh
        ci.Update(lset_p1,time_order=time_order)

        # re-compute the facets for stabilization:
        ba_facets = GetFacetsWithNeighborTypes(mesh,a=ci.GetElementsOfType(HASNEG),
                                                    b=ci.GetElementsOfType(IF))
        # re-evaluate the "active dofs" in the space time slab
        active_dofs = GetDofsOfElements(st_fes,ci.GetElementsOfType(HASNEG))

        # re-set definedonelements-markers according to new markings:
        for integrator in hasneg_integrators_a + hasneg_integrators_f:
            integrator.SetDefinedOnElements(ci.GetElementsOfType(HASNEG))
        for integrator in patch_integrators_a:
            integrator.SetDefinedOnElements(ba_facets)

        # assemble linear system
        a.Assemble()
        f.Assemble()

        # solve linear system
        inv = a.mat.Inverse(active_dofs,inverse="")
        gfu.vec.data =  inv * f.vec
        

        # evaluate upper trace of solution for
        #  * for error evaluation 
        #  * upwind-coupling to next time slab
        RestrictGFInTime(spacetime_gf=gfu,reference_time=1.0,space_gf=u_last)   

        # update time variable (float and ParameterCL)
        told = told + delta_t
        coef_told.Set(told)
        
        # compute error at end of time slab
        l2error = sqrt(Integrate(lset_neg_top,(u_exactL(told) -u_last)**2,mesh))
        # print time and error
        print("t = {0:10}, l2error = {1:20}".format(told,l2error),end="\n")
        assert(l2error < 0.085)

@pytest.mark.parametrize("i", [3,4,5])
def length_1D_Int_test():
    length = 1
    mesh = Make1DMesh(n=2**(i), mapping= lambda x : 2*length*x-length)
    
    r0 = 0.4
    r = sqrt(x**2)
    
    # level set
    levelset= r - r0
    
    fes1 = H1(mesh, order=1)

    lset_p1 = GridFunction(fes1)
    
    InterpolateToP1(levelset,lset_p1)
    val_vol = Integrate({ "levelset" : lset_p1, "domain_type" : NEG}, CoefficientFunction(1.0), mesh)
    print("Numerical value: ", val_vol)
    print("Error: ", abs(val_vol - 2*r0))
    assert( abs(val_vol - 2*r0) < 1e-10)

def area_of_a_circle_ST_error(n_steps = 8, i=1):
    length = 1
    mesh = Make1DMesh(n=2**(i), mapping= lambda x : 2*length*x-length)

    coef_told = Parameter(0)
    coef_delta_t = Parameter(0)
    tref = ReferenceTimeVariable()
    t = coef_told + coef_delta_t*tref
    
    r0 = 0.9
    r = sqrt(x**2+t**2)
    
    # level set
    levelset= r - r0
    
    time_order = 1
    fes1 = H1(mesh, order=1)
    tfe = ScalarTimeFE(time_order)
    st_fes = SpaceTimeFESpace(fes1,tfe)
    
    tend = 1
    delta_t = tend/n_steps
    coef_delta_t.Set(delta_t)
    told = 0

    lset_p1 = GridFunction(st_fes)
    
    sum_vol = 0
    sum_int = 0
    for i in range(n_steps):
        SpaceTimeInterpolateToP1(levelset,tref,lset_p1)
    
        val_vol = Integrate({ "levelset" : lset_p1, "domain_type" : NEG}, CoefficientFunction(1.0), mesh, time_order = time_order)
        val_int = Integrate({ "levelset" : lset_p1, "domain_type" : IF}, CoefficientFunction(1.0), mesh, time_order = time_order)
        #print(val_vol, val_int)
        sum_vol += val_vol*delta_t
        sum_int += val_int*delta_t
        
        told = told + delta_t
        coef_told.Set(told)

    print("SUM VOL: ", sum_vol)
    print("VOL: ", pi*r0**2/2)
    vol_err = abs(sum_vol - pi*r0**2/2)
    print("\t\tDIFF: ", vol_err)
    
    print("SUM INT: ", sum_int)
    print("AREA: ", 2*r0)
    int_err = abs(sum_int - 2*r0)
    print("\t\tDIFF: ",int_err)
    return (vol_err, int_err)

def test_spacetime_area_of_a_circle():
    l2errors_vol = []
    for i in range(6):
        (n_steps,i) =  (2**(i+2), i+1)
        (vol_err, int_err) = area_of_a_circle_ST_error(n_steps, i)
        l2errors_vol.append(vol_err)
        assert int_err < 1e-10
    
    print("L2 (VOL): ", l2errors_vol)
    eocs_vol = [log(l2errors_vol[i-1]/l2errors_vol[i])/log(2) for i in range(1,len(l2errors_vol))]
    print("EOCS (VOL): ", eocs_vol)
    avg = sum(eocs_vol)/len(eocs_vol)
    print("Average: ", avg)
    assert avg > 1.9
    
def area_of_a_sphere_ST_error(n_steps = 8, i=1, structured_mesh=False):
    if structured_mesh:
        length = 1
        mesh = MakeStructured2DMesh(quads=False,nx=2**(i),ny=2**(i),mapping= lambda x,y : (2*length*x-length,2*length*y-length))
    else:
        square = SplineGeometry()
        square.AddRectangle([-1,-1],[1,1])
        ngmesh = square.GenerateMesh(maxh=(1/2)**(i-1), quad_dominated=False)
        mesh = Mesh (ngmesh)

    coef_told = Parameter(0)
    coef_delta_t = Parameter(0)
    tref = ReferenceTimeVariable()
    t = coef_told + coef_delta_t*tref
    
    r0 = 0.9
    r = sqrt(x**2+y**2+t**2)
    
    # level set
    levelset= r - r0
    
    time_order = 1
    fes1 = H1(mesh, order=1)
    tfe = ScalarTimeFE(time_order)
    st_fes = SpaceTimeFESpace(fes1,tfe)
    
    tend = 1
    delta_t = tend/n_steps
    coef_delta_t.Set(delta_t)
    told = 0

    lset_p1 = GridFunction(st_fes)
    
    sum_vol = 0
    sum_int = 0
    for i in range(n_steps):
        SpaceTimeInterpolateToP1(levelset,tref,lset_p1)
    
        val_vol = Integrate({ "levelset" : lset_p1, "domain_type" : NEG}, CoefficientFunction(1.0), mesh, time_order = time_order)
        val_int = Integrate({ "levelset" : lset_p1, "domain_type" : IF}, CoefficientFunction(1.0), mesh, time_order = time_order)
        #print(val_vol, val_int)
        sum_vol += val_vol*delta_t
        sum_int += val_int*delta_t
        
        told = told + delta_t
        coef_told.Set(told)

    print("SUM VOL: ", sum_vol)
    print("VOL: ", 2/3*pi*r0**3)
    vol_err = abs(sum_vol - 2/3*pi*r0**3)
    print("\t\tDIFF: ", vol_err)
    
    print("SUM INT: ", sum_int)
    print("AREA: ", 0.5*pi**2*r0**2)
    int_err = abs(sum_int - 0.5*pi**2*r0**2)
    print("\t\tDIFF: ",int_err)
    return (vol_err, int_err)

def area_of_a_sphere_ST_error_ho(n_steps = 8, i=3, structured_mesh=False, k = 3):
    if structured_mesh:
        length = 1
        mesh = MakeStructured2DMesh(quads=False,nx=2**(i),ny=2**(i),mapping= lambda x,y : (2*length*x-length,2*length*y-length))
    else:
        square = SplineGeometry()
        square.AddRectangle([-1,-1],[1,1])
        ngmesh = square.GenerateMesh(maxh=(1/2)**(i-1), quad_dominated=False)
        mesh = Mesh (ngmesh)

    coef_told = Parameter(0)
    coef_delta_t = Parameter(0)
    tref = ReferenceTimeVariable()
    t = coef_told + coef_delta_t*tref
    
    r0 = 0.9
    r = sqrt(x**2+y**2+t**2)
    
    # level set
    levelset= r - r0
    
    lsetadap = LevelSetMeshAdaptation_Spacetime(mesh, order_space=k, order_time=k, threshold=0.5, discontinuous_qn=True)
    
    time_order = k
    
    tend = r0/2
    delta_t = tend/n_steps
    coef_delta_t.Set(delta_t)
    told = 0
    
    dQ = delta_t * dCut(lsetadap.levelsetp1[INTERVAL], NEG, time_order=time_order, order = k,
                        deformation=lsetadap.deformation[INTERVAL])
    dG = delta_t * dCut(lsetadap.levelsetp1[INTERVAL], IF, time_order=time_order, order = k,
                        deformation=lsetadap.deformation[INTERVAL])
    
    sum_vol = 0
    sum_int = 0
    for i in range(n_steps):
        lsetadap.CalcDeformation(levelset)
    
        val_vol = Integrate( CF(1.)*dQ, mesh)
        val_int = Integrate( CF(1.)*dG, mesh)
        #print(val_vol, val_int)
        sum_vol += val_vol
        sum_int += val_int
        
        told = told + delta_t
        coef_told.Set(told)

    print("SUM VOL: ", sum_vol)
    print("VOL: ", 11/24*pi*r0**3)
    vol_err = abs(sum_vol - 11/24*pi*r0**3)
    print("\t\tDIFF: ", vol_err)
    
    print("SUM INT: ", sum_int)
    print("AREA: ", pi/12*r0**2*(3*sqrt(3) + 2*pi))
    int_err = abs(sum_int - pi/12*r0**2*(3*sqrt(3) + 2*pi))
    print("\t\tDIFF: ",int_err)
    return (vol_err, int_err)

@pytest.mark.parametrize("structured", [True, False])
def test_spacetime_area_of_a_sphere(structured):
    
    l2errors_vol = []
    l2errors_int = []
    for i in range(6):
        (n_steps,i) =  (2**(i+2), i+1)
        (vol_err, int_err) = area_of_a_sphere_ST_error(n_steps, i, structured)
        l2errors_vol.append(vol_err)
        l2errors_int.append(int_err)
    
    print("L2 (VOL): ", l2errors_vol)
    eocs_vol = [log(l2errors_vol[i-1]/l2errors_vol[i])/log(2) for i in range(1,len(l2errors_vol))]
    print("EOCS (VOL): ", eocs_vol)
    avg = sum(eocs_vol)/len(eocs_vol)
    print("Average: ", avg)
    assert avg > 1.9
    
    print("L2 (INT): ", l2errors_int)
    eocs_int = [log(l2errors_int[i-1]/l2errors_int[i])/log(2) for i in range(1,len(l2errors_int))]
    print("EOCS (INT): ", eocs_int)
    avg = sum(eocs_int)/len(eocs_int)
    print("Average: ", avg)
    assert avg > 1.9

@pytest.mark.parametrize("structured", [True, False])
@pytest.mark.parametrize("k", [2,3,4,5,6,7,8])
def test_spacetime_area_of_a_sphere_ho(structured, k):
    
    l2errors_vol = []
    l2errors_int = []
    for i in range(6):
        (n_steps,i) =  (2**(i+1), i+1)
        (vol_err, int_err) = area_of_a_sphere_ST_error_ho(n_steps, i, structured, k)
        l2errors_vol.append(vol_err)
        l2errors_int.append(int_err)
        if vol_err < 1e-9 or int_err < 1e-9:
            break
    
    print("L2 (VOL): ", l2errors_vol)
    eocs_vol = [log(l2errors_vol[i-1]/l2errors_vol[i])/log(2) for i in range(1,len(l2errors_vol))]
    print("EOCS (VOL): ", eocs_vol)
    avg = sum(eocs_vol)/len(eocs_vol)
    print("Average: ", avg)
    assert avg > k+0.8
    
    print("L2 (INT): ", l2errors_int)
    eocs_int = [log(l2errors_int[i-1]/l2errors_int[i])/log(2) for i in range(1,len(l2errors_int))]
    print("EOCS (INT): ", eocs_int)
    avg = sum(eocs_int)/len(eocs_int)
    print("Average: ", avg)
    assert avg > k+0.8

#ngsxfemglobals.do_naive_timeint = True
#ngsxfemglobals.naive_timeint_order = 2
#ngsxfemglobals.naive_timeint_subdivs = 1
#test_spacetime_area_of_a_sphere_ho(False, 2)

def area_of_a_hypersphere_ST_error(n_steps = 64, i=1, structured_mesh= True):
    if structured_mesh:
        length = 1
        mesh = MakeStructured3DMesh(hexes=False,nx=2**(i),ny=2**(i), nz=2**(i),mapping= lambda x,y,z : (2*length*x-length,2*length*y-length, 2*length*z - length))
    else:
        cube = CSGeometry()
        cube.Add (OrthoBrick(Pnt(-1,-1,-1), Pnt(1,1,1)))
        ngmesh = cube.GenerateMesh(maxh=(1/2)**(i-1), quad_dominated=False)
        mesh = Mesh (ngmesh)

    coef_told = Parameter(0)
    coef_delta_t = Parameter(0)
    tref = ReferenceTimeVariable()
    t = coef_told + coef_delta_t*tref
    
    r0 = 0.9
    r = sqrt(x**2+y**2+z**2+t**2)
    
    # level set
    levelset= r - r0
    
    time_order = 1
    fes1 = H1(mesh, order=1)
    tfe = ScalarTimeFE(time_order)
    st_fes = SpaceTimeFESpace(fes1,tfe)

    tend = 1
    delta_t = tend/n_steps
    coef_delta_t.Set(delta_t)
    told = 0
    
    lset_p1 = GridFunction(st_fes)
    
    sum_vol = 0
    sum_int = 0
    for i in range(n_steps):
        SpaceTimeInterpolateToP1(levelset,tref,lset_p1)
    
        val_vol = Integrate({ "levelset" : lset_p1, "domain_type" : NEG}, CoefficientFunction(1.0), mesh, time_order = time_order)
        val_int = Integrate({ "levelset" : lset_p1, "domain_type" : IF}, CoefficientFunction(1.0), mesh, time_order = time_order)
        #print(val_vol, val_int)
        sum_vol += val_vol*delta_t
        sum_int += val_int*delta_t
        
        told = told + delta_t
        coef_told.Set(told)

    print("SUM VOL: ", sum_vol)
    print("VOL: ", pi**2/4*r0**4)
    vol_err = abs(sum_vol - pi**2/4*r0**4)
    print("\t\tDIFF: ", vol_err)
    
    print("SUM INT: ", sum_int)
    print("AREA: ", 8/3*pi*r0**3)
    int_err = abs(sum_int - 8/3*pi*r0**3)
    print("\t\tDIFF: ", int_err)
    return (vol_err, int_err)

@pytest.mark.parametrize("structured", [True, False])
def test_spacetime_area_of_a_hypersphere(structured):
    l2errors_vol = []
    l2errors_int = []
    for i in range(3):
        (n_steps,i) =  (2**(i+3), i+2)
        (vol_err, int_err) = area_of_a_hypersphere_ST_error(n_steps, i, structured)
        l2errors_vol.append(vol_err)
        l2errors_int.append(int_err)
    
    print("L2 (VOL): ", l2errors_vol)
    eocs_vol = [log(l2errors_vol[i-1]/l2errors_vol[i])/log(2) for i in range(1,len(l2errors_vol))]
    print("EOCS (VOL): ", eocs_vol)
    avg = sum(eocs_vol)/len(eocs_vol)
    print("Average: ", avg)
    assert avg > 1.9
    
    print("L2 (INT): ", l2errors_int)
    eocs_int = [log(l2errors_int[i-1]/l2errors_int[i])/log(2) for i in range(1,len(l2errors_int))]
    print("EOCS (INT): ", eocs_int)
    avg = sum(eocs_int)/len(eocs_int)
    print("Average: ", avg)
    assert avg > 1.9

@pytest.mark.skip(reason="function is tested in test_utils")
def test_spacetime_spaceP4_timeDGP4():
    ngsglobals.msg_level = 1

    # -------------------------------- PARAMETERS ---------------------------------
    # DISCRETIZATION PARAMETERS:

    # Parameter for refinement study:
    i = 2
    n_steps = 2**i
    space_refs = i

    # Polynomial order in time
    k_t = 4
    # Polynomial order in space
    k_s = k_t
    # Polynomial order in time for level set approximation
    lset_order_time = k_t
    # Integration order in time
    time_order = 2 * k_t
    # Time stepping parameters
    tstart = 0
    tend = 0.5
    delta_t = (tend - tstart) / n_steps
    maxh = 0.5
    # Ghost-penalty parameter
    gamma = 0.05
    # Map from reference time to physical time
    told = Parameter(tstart)
    t = told + delta_t * tref

    # PROBLEM SETUP:

    # Outer domain:
    rect = SplineGeometry()
    rect.AddRectangle([-0.6, -1], [0.6, 1])

    # Level set geometry
    # Radius of disk (the geometry)
    R = 0.5
    # Position shift of the geometry in time
    rho = (1 / (pi)) * sin(2 * pi * t)
    # Convection velocity:
    w = CoefficientFunction((0, rho.Diff(t)))
    # Level set
    r = sqrt(x**2 + (y - rho)**2)
    levelset = r - R

    # Diffusion coefficient
    alpha = 1
    # Solution
    u_exact = cos(pi * r / R) * sin(pi * t)
    # R.h.s.
    coeff_f = (u_exact.Diff(t)
            - alpha * (u_exact.Diff(x).Diff(x) + u_exact.Diff(y).Diff(y))
            + w[0] * u_exact.Diff(x) + w[1] * u_exact.Diff(y)).Compile()

    # ----------------------------------- MAIN ------------------------------------
    ngmesh = rect.GenerateMesh(maxh=maxh, quad_dominated=False)
    for j in range(space_refs):
        ngmesh.Refine()
    mesh = Mesh(ngmesh)

    # Spatial FESpace for solution
    fes1 = H1(mesh, order=k_s, dgjumps=True)
    # Time finite element (nodal!)
    tfe = ScalarTimeFE(k_t)
    # (Tensor product) space-time finite element space
    st_fes = tfe * fes1

    # Space time version of Levelset Mesh Adapation object. Also offers integrator
    # helper functions that involve the correct mesh deformation
    lsetadap = LevelSetMeshAdaptation_Spacetime(mesh, order_space=k_s,
                                                order_time=lset_order_time,
                                                threshold=0.5,
                                                discontinuous_qn=True)

    gfu = GridFunction(st_fes)
    u_last = CreateTimeRestrictedGF(gfu, 1)

    scene = DrawDC(lsetadap.levelsetp1[TOP], u_last, 0, mesh, "u_last",
                deformation=lsetadap.deformation[TOP])

    u, v = st_fes.TnT()
    h = specialcf.mesh_size

    ba_facets = BitArray(mesh.nfacet)
    ci = CutInfo(mesh, time_order=0)

    dQ = delta_t * dCut(lsetadap.levelsetp1[INTERVAL], NEG, time_order=time_order,
                        deformation=lsetadap.deformation[INTERVAL],
                        definedonelements=ci.GetElementsOfType(HASNEG))
    dOmold = dCut(lsetadap.levelsetp1[BOTTOM], NEG,
                deformation=lsetadap.deformation[BOTTOM],
                definedonelements=ci.GetElementsOfType(HASNEG), tref=0)
    dOmnew = dCut(lsetadap.levelsetp1[TOP], NEG,
                deformation=lsetadap.deformation[TOP],
                definedonelements=ci.GetElementsOfType(HASNEG), tref=1)
    dw = delta_t * dFacetPatch(definedonelements=ba_facets, time_order=time_order,
                            deformation=lsetadap.deformation[INTERVAL])


    def dt(u):
        return 1.0 / delta_t * dtref(u)


    a = RestrictedBilinearForm(st_fes, "a", check_unused=False,
                            element_restriction=ci.GetElementsOfType(HASNEG),
                            facet_restriction=ba_facets)
    a += v * (dt(u) - dt(lsetadap.deform) * grad(u)) * dQ
    a += (alpha * InnerProduct(grad(u), grad(v))) * dQ
    a += (v * InnerProduct(w, grad(u))) * dQ
    a += u * v * dOmold
    a += h**(-2) * (1 + delta_t / h) * gamma * \
        (u - u.Other()) * (v - v.Other()) * dw

    f = LinearForm(st_fes)
    f += coeff_f * v * dQ
    f += u_last * v * dOmold

    # Set initial values
    u_last.Set(fix_tref(u_exact, 0))
    # Project u_last at the beginning of each time step
    lsetadap.ProjectOnUpdate(u_last)

    while tend - told.Get() > delta_t / 2:
        lsetadap.CalcDeformation(levelset)

        # Update markers in (space-time) mesh
        ci.Update(lsetadap.levelsetp1[INTERVAL], time_order=0)

        # re-compute the facets for stabilization:
        ba_facets[:] = GetFacetsWithNeighborTypes(mesh,
                                                a=ci.GetElementsOfType(HASNEG),
                                                b=ci.GetElementsOfType(IF))
        active_dofs = GetDofsOfElements(st_fes, ci.GetElementsOfType(HASNEG))

        a.Assemble(reallocate=True)
        f.Assemble()

        # Solve linear system
        inv = a.mat.Inverse(active_dofs, inverse="")
        gfu.vec.data = inv * f.vec.data

        # Evaluate upper trace of solution for
        #  * for error evaluation
        #  * upwind-coupling to next time slab
        RestrictGFInTime(spacetime_gf=gfu, reference_time=1.0, space_gf=u_last)

        # Compute error at final time
        l2error = sqrt(Integrate((u_exact - u_last)**2 * dOmnew, mesh))

        # Update time variable (ParameterCL)
        told.Set(told.Get() + delta_t)
        print("\rt = {0:12.9f}, L2 error = {1:12.9e}".format(told.Get(), l2error))
    assert(l2error < 1e-2)
    return l2error



"""
In this example we solve an *unfitted* Stokes interface problem with a high order isoparametric
unfitteddiscretization method. 
This file is based on the discretizations in cutfem.py, nxfem.py and nxfem_higher_order.py. 

    domain: 
    -------
    The domain is [-1,1]^2 while the interface is described by a level set function ( circile with
    radius R=2/3).

    PDE problem:
    ------------
    domain equations for velocity (u1,u2) and pressure p:
     - alpha_i (u1_xx + u1_yy) + p_x = rho_i g1 in subdomain i, i=1,2,
     - alpha_i (u2_xx + u2_yy) + p_y = rho_i g2 in subdomain i, i=1,2,
                         u1_x + u2_y = 0        in subdomain i, i=1,2,
    interface conditions:
                                [u] =    0 on interface (continuity across the interface    ),
             [-alpha · du/dn + p·n] =    f on interface (conservation of the (momentum) flux),
                                 u  =  u_D on domain boundary.

    The r.h.s. term (g1,g2) corresponds to gravity, the term f is surface tension force (here f =
    kappa · n where kappa is the mean curvature (1/R)). The Dirichlet data is chosen according to a
    manufactured solution introduced in [1]
    which allows us to measure errors after the computation of a discrete solution.
    The coefficients alpha are domain-wise constants which are different in the two subdomains.
    

    discretization:
    ---------------
    Finite element space:
    As in nxfem.py but for every velocity component and the pressure

    Variational formulation:
    We use a Nitsche formulation which involves averages of the fluxes and jumps of the solution
    across the interface [2]. For the average we use the Hansbo-choice [3] where the average is
    adjusted to the local cut configuration in order to ensure stability of the resulting viscosity
    formulation. 
    To ensure inf-sup for the velocity-pressure space we add a ghost penalty stabilization on the
    pressure space, cf. [2]. 

    Surface tension:  
    In this example we prescribe the surface tension analytically, i.e. we circumvent approximating
    the mean curvature from the level set function. 

    implementational aspects:
    ---------------
    Geometry approximation:
    As in nxfem_higher_order.py

    Ghost penalty stabilization:
    The edge-based stabilizations require different couplings (compared to standard
    discretizations). To add these to the sparsity pattern we have to add the "dgjumps" flags which
    prepares the sparse matrix for the corresponding needed couplings.


    linear systems:
    ---------------
    A (sparse) direct solver is applied to solve the arising linear systems.

    literature:
    -----------
    [1]: M.Kirchhart, S.Groß, A.Reusken, Analysis of an XFEM discretization for Stokes interface
    problems, SISC, 2016
    [2]: P.Lederer, C.-M.Pfeiler, C.Wintersteiger, C.Lehrenfeld, Higher order unfitted fem for
    stokes interface problems. PAMM, 2016 
    [3]: A.Hansbo, P.Hansbo, An unfitted finite element method, based on Nitsche's method, for
    elliptic interface problems, Comp. Meth. Appl. Mech. Eng., 2002

"""

from netgen.geom2d import SplineGeometry
from netgen.meshing import MeshingParameters
from ngsolve import *
from xfem import *
# For LevelSetAdaptationMachinery
from xfem.lsetcurv import *
from math import pi
# 2D: circle configuration
square = SplineGeometry()
square.AddRectangle([-1,-1],[1,1],bcs=[1,2,3,4])
mesh = Mesh (square.GenerateMesh(maxh=4, quad_dominated=False))
for i in range(3):
    mesh.Refine()

order = 2 #=k in P(k)P(k-1)-Taylor-Hood pair
    
mu1 = 1.0
mu2 = 10.0
mu = [mu1,mu2]

R = 2.0/3.0
aneg = 1.0/mu1
apos = 1.0/mu2 + (1.0/mu1 - 1.0/mu2)*exp(x*x+y*y-R*R)
gammaf = 0.5 # surface tension = pressure jump
#q = gammaf - pi*R*R/4.0*gammaf

functions = {
           "Levelset" : sqrt(x*x+y*y)-R,
           "SourceInnerX" : (exp(-1*(x*x+y*y))*((-8*y)+(4*x*x*y)+(4*y*y*y))+3*x*x),
           "SourceOuterX" : (exp(-1*(x*x+y*y))*((-8*y)+(4*x*x*y)+(4*y*y*y))+3*x*x),
           "SourceInnerY" : (exp(-1*(x*x+y*y))*((-4*x*x*x)+(8*x)-(4*x*y*y))),
           "SourceOuterY" : (exp(-1*(x*x+y*y))*((-4*x*x*x)+(8*x)-(4*x*y*y))),
           "SolutionOuterVelX" : (apos*exp(-1.0 *( x * x + y * y)) * -1.0 * y),
           "SolutionOuterVelY" : (apos*exp(-1.0 *( x * x + y * y)) * x),
           "SolutionInnerVelX" : (aneg*exp(-1.0 *( x * x + y * y)) * -1.0 * y),
           "SolutionInnerVelY" : (aneg*exp(-1.0 *( x * x + y * y)) * x),
           "SolutionOuterVelX_DX" : (2.0*x*y/mu2*exp(-1.0 *( x * x + y * y))),
           "SolutionOuterVelX_DY" : ((2.0*y*y/mu2-apos)*exp(-1.0 *( x * x + y * y))),
           "SolutionOuterVelY_DX" : ((apos - 2*x*x / mu2)*exp(-1.0 *( x * x + y * y))),
           "SolutionOuterVelY_DY" : (-2.0*x*y/mu2*exp(-1.0 *( x * x + y * y))),
           "SolutionInnerVelX_DX" : (aneg*2*x*y*exp(-1.0 *( x * x + y * y))),
           "SolutionInnerVelX_DY" : (aneg*(2*y*y-1)*exp(-1.0 *( x * x + y * y))),
           "SolutionInnerVelY_DX" : (aneg*(1-2*x*x)*exp(-1.0 *( x * x + y * y))),
           "SolutionInnerVelY_DY" : (aneg*(-2)*x*y*exp(-1.0 *( x * x + y * y))),
           "SolutionInnerPressure" : (x*x*x),# + q),
           "SolutionOuterPressure" : (x*x*x - gammaf), #(pi*R*R/4.0*gammaf)),
}

coef_g = [ CoefficientFunction((functions["SourceInnerX"],functions["SourceInnerY"])),
           CoefficientFunction((functions["SourceOuterX"],functions["SourceOuterY"])) ]

vel_sol_neg = CoefficientFunction((functions["SolutionInnerVelX"],functions["SolutionInnerVelY"]))
vel_sol_pos = CoefficientFunction((functions["SolutionOuterVelX"],functions["SolutionOuterVelY"]))

grad_vel_sol_neg = CoefficientFunction((functions["SolutionInnerVelX_DX"],functions["SolutionInnerVelX_DY"],functions["SolutionInnerVelY_DX"],functions["SolutionInnerVelY_DY"]))
grad_vel_sol_pos = CoefficientFunction((functions["SolutionOuterVelX_DX"],functions["SolutionOuterVelX_DY"],functions["SolutionOuterVelY_DX"],functions["SolutionOuterVelY_DY"]))

pres_sol_neg = functions["SolutionInnerPressure"]
pres_sol_pos = functions["SolutionOuterPressure"]

# stabilization parameter for ghost-penalty
gamma_stab = 0.05
# stabilization parameter for Nitsche
lambda_nitsche  = 0.5 * (mu[0]+mu[1]) * 20 * order * order

levelset = functions["Levelset"]

lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=10.5, discontinuous_qn=True)
deformation = lsetmeshadap.CalcDeformation(levelset)
lsetp1 = lsetmeshadap.lset_p1
ci = CutInfo(mesh,lsetp1)

Vh = H1(mesh, order=order, dirichlet=[1,2,3,4])
Vhx = XFESpace(Vh, ci)
VhG = FESpace([Vh, Vhx])
Qh = H1(mesh, order=order-1)
Qhx = XFESpace(Qh, ci)
QhG = FESpace([Qh, Qhx])
N = FESpace("number",mesh)

WhG = FESpace([VhG,VhG,QhG,N],flags = {"dgjumps" : True})

gfup = GridFunction(WhG)
gfu1 = gfup.components[0]
gfu2 = gfup.components[1]
gfp = gfup.components[2]
u1_coef = gfu1.components[0] + IfPos(lsetp1, pos(gfu1.components[1]), neg(gfu1.components[1]))
u2_coef = gfu2.components[0] + IfPos(lsetp1, pos(gfu2.components[1]), neg(gfu2.components[1]))

vel_neg = CoefficientFunction((gfu1.components[0]+neg(gfu1.components[1]),
                               gfu2.components[0]+neg(gfu2.components[1])))
vel_pos = CoefficientFunction((gfu1.components[0]+pos(gfu1.components[1]),
                               gfu2.components[0]+pos(gfu2.components[1])))
vel_coef = CoefficientFunction((u1_coef,u2_coef))

grad_vel_neg = CoefficientFunction((grad(gfu1.components[0])+neg_grad(gfu1.components[1]),
                                    grad(gfu2.components[0])+neg_grad(gfu2.components[1])))
grad_vel_pos = CoefficientFunction((grad(gfu1.components[0])+pos_grad(gfu1.components[1]),
                                    grad(gfu2.components[0])+pos_grad(gfu2.components[1])))


pres_neg = gfp.components[0]+neg(gfp.components[1])
pres_pos = gfp.components[0]+pos(gfp.components[1])
pres_coef = gfp.components[0] + IfPos(lsetp1, pos(gfp.components[1]), neg(gfp.components[1]))

gfp = gfup.components[2]

n_outer = specialcf.normal(mesh.dim)
h = specialcf.mesh_size   

lset_neg = { "levelset" : lsetp1, "domain_type" : NEG, "subdivlvl" : 0}
lset_pos = { "levelset" : lsetp1, "domain_type" : POS, "subdivlvl" : 0}
lset_if  = { "levelset" : lsetp1, "domain_type" : IF , "subdivlvl" : 0}

# element, facet and dof marking w.r.t. boundary approximation with lsetp1:
hasneg = ci.GetElementsOfType(HASNEG)
haspos = ci.GetElementsOfType(HASPOS)
hasif = ci.GetElementsOfType(IF)
ba_facets = [GetFacetsWithNeighborTypes(mesh,a=hasneg,b=hasif),GetFacetsWithNeighborTypes(mesh,a=haspos,b=hasif)]

n_lset = 1.0/Norm(grad(lsetp1)) * grad(lsetp1)
kappa = [CutRatioGF(ci),1.0-CutRatioGF(ci)]
           
a = BilinearForm(WhG,symmetric=False)
f = LinearForm(WhG)
            
(u1s,u1x),(u2s,u2x),(ps,px),n = WhG.TrialFunction()
(v1s,v1x),(v2s,v2x),(qs,qx),m = WhG.TestFunction()

u1 = [u1s + op(u1x) for op in [neg,pos]] 
u2 = [u2s + op(u2x) for op in [neg,pos]]
u = [CoefficientFunction((u1[i],u2[i])) for i in range(2)]
p = [ps + op(px) for op in [neg,pos]] 
dpdn = [grad(ps)*n_outer + op(px)*n_outer for op in [neg_grad,pos_grad]] 
dpdn_Other = [grad(ps.Other())*n_outer + op(px.Other())*n_outer for op in [neg_grad,pos_grad]] 
dpdn_jump = [dpdn[i] - dpdn_Other[i] for i in range(2)] 
gradu1 = [grad(u1s) + op(u1x) for op in [neg_grad,pos_grad]]
gradu2 = [grad(u2s) + op(u2x) for op in [neg_grad,pos_grad]]
divu = [gradu1[i][0]+gradu2[i][1] for i in range(2)]
Du = [CoefficientFunction((2*gradu1[i][0],gradu2[i][0]+gradu1[i][1],gradu2[i][0]+gradu1[i][1],2*gradu2[i][1]),dims = (2,2)) for i in range(2)]
sigmaupn = [ - mu[i] * (Du[i] * n_lset) + p[i] * n_lset for i in range(2)] 
average_flux_u = kappa[0] * sigmaupn[0] + kappa[1] * sigmaupn[1]

v1 = [v1s + op(v1x) for op in [neg,pos]] 
v2 = [v2s + op(v2x) for op in [neg,pos]] 
v = [CoefficientFunction((v1[i],v2[i])) for i in range(2)]
q = [qs + op(qx) for op in [neg,pos]] 
dqdn = [grad(qs)*n_outer + op(qx)*n_outer for op in [neg_grad,pos_grad]] 
dqdn_Other = [grad(qs.Other())*n_outer + op(qx.Other())*n_outer for op in [neg_grad,pos_grad]] 
dqdn_jump = [dqdn[i] - dqdn_Other[i] for i in range(2)] 
gradv1 = [grad(v1s) + op(v1x) for op in [neg_grad,pos_grad]]
gradv2 = [grad(v2s) + op(v2x) for op in [neg_grad,pos_grad]]
divv = [gradv1[i][0]+gradv2[i][1] for i in range(2)]
Dv = [CoefficientFunction((2*gradv1[i][0],gradv2[i][0]+gradv1[i][1],gradv2[i][0]+gradv1[i][1],2*gradv2[i][1]),dims = (2,2)) for i in range(2)]
sigmavqn = [ - mu[i] * Dv[i] * n_lset + q[i] * n_lset for i in range(2)]
average_flux_v = kappa[0] * sigmavqn[0] + kappa[1] * sigmavqn[1]
average_inv_v = - kappa[1] * v[0] - kappa[0] * v[1]

# Viscosity term
a += SymbolicBFI(lset_neg, form = 0.5 * mu1 * InnerProduct(Du[0],Dv[0]) )
a += SymbolicBFI(lset_pos, form = 0.5 * mu2 * InnerProduct(Du[1],Dv[1]) )
a += SymbolicBFI(lset_if,  form = InnerProduct( average_flux_u, v[0]-v[1]))
a += SymbolicBFI(lset_if,  form = InnerProduct( average_flux_v, u[0]-u[1]))
a += SymbolicBFI(lset_if,  form = lambda_nitsche/h * InnerProduct(u[0]-u[1],v[0]-v[1]))
# pressure stuff
a += SymbolicBFI(lset_neg, form = - divu[0] * q[0] - divv[0] * p[0]
                                  + n * q[0] + m * p[0])
a += SymbolicBFI(lset_pos, form = - divu[1] * q[1] - divv[1] * p[1])
#                                  + n * q[1] + m * p[1])

f += SymbolicLFI(lset_if,  form = gammaf * InnerProduct(average_inv_v,n_lset))
f += SymbolicLFI(lset_neg, form = coef_g[0] * v[0])
f += SymbolicLFI(lset_pos, form = coef_g[1] * v[1])


# # ghost penalty terms:
for i in range(2):
    a += SymbolicBFI(form = - gamma_stab * h*h*h* dpdn_jump[i]*dqdn_jump[i],VOL_or_BND = VOL, skeleton=True, definedonelements=ba_facets[i])

# apply mesh adaptation    
mesh.SetDeformation(deformation)

a.Assemble()
f.Assemble()

gfu1.components[0].Set(functions["SolutionOuterVelX"])
gfu2.components[0].Set(functions["SolutionOuterVelY"])
# Solve linear system
f.vec.data -= a.mat * gfup.vec
gfup.vec.data += a.mat.Inverse(WhG.FreeDofs()) * f.vec
              
# #measure the error

vl2error = sqrt( Integrate(lset_neg,InnerProduct(vel_neg-vel_sol_neg,vel_neg-vel_sol_neg),mesh=mesh)
                 +Integrate(lset_pos,InnerProduct(vel_pos-vel_sol_pos,vel_pos-vel_sol_pos),mesh=mesh) )
print("L2 Error of velocity: {0}".format(vl2error))

vh1error = sqrt( Integrate(lset_neg,InnerProduct(grad_vel_neg-grad_vel_sol_neg,grad_vel_neg-grad_vel_sol_neg),mesh=mesh)
                 +Integrate(lset_pos,InnerProduct(grad_vel_pos-grad_vel_sol_pos,grad_vel_pos-grad_vel_sol_pos),mesh=mesh) )
print("H1 Error of velocity: {0}".format(vh1error))

pl2error = sqrt(  Integrate(lset_neg,(pres_neg-pres_sol_neg)*(pres_neg-pres_sol_neg),mesh=mesh)
                 +Integrate(lset_pos,(pres_pos-pres_sol_pos)*(pres_pos-pres_sol_pos),mesh=mesh) )
print("L2 Error of pressure: {0}".format(pl2error))

# # unset mesh adaptation
mesh.UnsetDeformation()

# #visualization:
Draw(deformation,mesh,"deformation")
Draw(levelset,mesh,"levelset")
Draw(lsetp1,mesh,"lsetp1")
Draw(pres_coef,mesh,"pressure")
Draw(vel_coef,mesh,"vel")



"""
Example of a triangle described by three level set functions
"""

from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.mlset import *

# Background mesh
square = SplineGeometry()
square.AddRectangle([-1,-0.5], [1,1.5], bc=1)
mesh = Mesh (square.GenerateMesh(maxh=1, quad_dominated=False))

# Level sets
lsetp1_a = GridFunction(H1(mesh,order=1))
lsetp1_b = GridFunction(H1(mesh,order=1))
lsetp1_c = GridFunction(H1(mesh,order=1))

InterpolateToP1(     y-1,lsetp1_a)
InterpolateToP1( 2*x-y  ,lsetp1_b)
InterpolateToP1(-2*x-y  ,lsetp1_c)

Draw (lsetp1_a, mesh, "lset_a")
Draw (lsetp1_b, mesh, "lset_b")
Draw (lsetp1_c, mesh, "lset_c")

Draw (IfPos(lsetp1_a,0,1) * IfPos(lsetp1_b,0,1) * IfPos(lsetp1_c,0,1), mesh, "lset_mult")

# Integration over multiple level sets
triangle = DomainTypeArray([(NEG,NEG,NEG)])


area = Integrate(levelset_domain={"levelset": (lsetp1_a,lsetp1_b,lsetp1_c),
                                  "domain_type": triangle.as_list},
                 mesh=mesh, cf=1, order=0)
print("area = {:10.8f}".format(area))                     
print("area error = {:4.3e}".format(abs(area - 0.5)))  
assert abs(area-0.5) < 1e-12

area_invert = Integrate(levelset_domain={"levelset": (lsetp1_a,lsetp1_b,lsetp1_c), 
                                         "domain_type": (~triangle).as_list},
                        mesh=mesh, cf=1, order=0)
print("area_invert = {:10.8f}".format(area_invert))                     
print("area_invert error = {:4.3e}".format(abs(area_invert - 3.5)))  
assert abs(area_invert-3.5) < 1e-12
                   
perimeter = Integrate(levelset_domain={"levelset" : (lsetp1_a,lsetp1_b,lsetp1_c), 
                                       "domain_type": triangle.Boundary().as_list},
                    mesh=mesh, cf=1, order=0)
print("perimeter =", perimeter)                     
print("perimeter error =", abs(perimeter - 1 - sqrt(5)))                     
assert abs(perimeter - 1 - sqrt(5)) < 1e-12


all_points = Integrate(levelset_domain={"levelset": (lsetp1_a,lsetp1_b,lsetp1_c), 
                                        "domain_type": triangle.Boundary().Boundary().as_list},
                        mesh=mesh, cf=y, order=0)
print("all_points =", all_points)                     
print("all_points error =", abs(all_points - 2))
assert abs(all_points - 2) < 1e-12




"""
This is a simple demonstration on how to use the SpaceTime VTK output
for levelset-based moving domain problems in 2D. A circle is travelling
around in a squared background domain. A piecewise linear-in-space
approximation of the level set geometry is taken as an initial
approximation. A space-time mesh deformation is then applied to recover
higher order geometrical accuracy also in space-time. In this example
only a few time steps are carried out and in each time step a VTK output
is generated which allows to visualize the space-time geometries.
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.geom2d import SplineGeometry
from ngsolve import *
from ngsolve.solvers import *
from ngsolve.internal import *

from xfem import *
from xfem.lset_spacetime import *

from math import pi

ngsglobals.msg_level = 1

# -------------------------------- PARAMETERS ---------------------------------
# Half side-length of background square domain
A = 1.25
# Mesh size
maxh = 0.3

# Radius of disk (the geometry)
r0 = 0.5

# Polynomial order in time
k_t = 2
# Polynomial order in space
k_s = 2
# Polynomial order in time for level set approximation
lset_order_time = 1
# Integration order in time
time_order = 2

# End time
tend = 1
# Time-step
delta_t = tend / 4

# ----------------------------------- MAIN ------------------------------------
square = SplineGeometry()
square.AddRectangle([-A, -A], [A, A])
ngmesh = square.GenerateMesh(maxh=maxh, quad_dominated=False)
mesh = Mesh(ngmesh)

# Expression for the time variable:
told = Parameter(0)
t = told + delta_t * tref


# The data:
def x0(t):
    return r0 * cos(pi * t)


def y0(t):
    return r0 * sin(pi * t)


# Level set = x - t
levelset = sqrt((x - x0(t))**2 + (y - y0(t))**2) - 0.4


# Spatial FESpace for solution
fesp1 = H1(mesh, order=1, dgjumps=True)

# Time finite element (nodal!)
tfe = ScalarTimeFE(k_t)
# Space-time finite element space
st_fes = tfe * fesp1

# Unfitted heat equation example
lset_p1 = GridFunction(st_fes)

lset_adap_st = LevelSetMeshAdaptation_Spacetime(mesh, order_space=k_s,
                                                order_time=lset_order_time,
                                                threshold=0.1,
                                                discontinuous_qn=True)

ci = CutInfo(mesh, time_order=0)

vtk_out = [levelset, lset_p1, CoefficientFunction((lset_adap_st.deform[0],
                                                   lset_adap_st.deform[1], 0)),
           BitArrayCF(ci.GetElementsOfType(IF))]
vtk_out_names = ["levelset", "lset_p1", "deform", "cutelements"]

vtk = SpaceTimeVTKOutput(ma=mesh, coefs=vtk_out, names=vtk_out_names,
                         filename="spacetime_vtk_", subdivision_x=3,
                         subdivision_t=3)

while tend - told.Get() > delta_t / 2:
    SpaceTimeInterpolateToP1(levelset, tref, lset_p1)
    dfm = lset_adap_st.CalcDeformation(levelset, tref)
    ci.Update(lset_p1, time_order=0)

    vtk.Do(t_start=told.Get(), t_end=told.Get() + delta_t)
    told.Set(told.Get() + delta_t)



# unfitted Heat equation with Neumann b.c.
# higher order geometry approximation with space-time mesh transformation
from ngsolve import *
from time import sleep
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry
from netgen.meshing import MeshingParameters

from ngsolve.internal import *
from ngsolve.solvers import *
from xfem import *
from math import pi
import sys

from xfem.lset_spacetime import *

ngsglobals.msg_level = 1

i = 5
gamma = 0.5

if hasattr(sys, 'argv') and len(sys.argv) == 5 and sys.argv[1] == "i" and sys.argv[3] == "stab":
    i = int(sys.argv[2])
    gamma = float(sys.argv[4])
    print("Sys argv :", sys.argv)
    print("Loading manual val for i, gamma: ", i, gamma)
else:
    print("Loading default val for i, gamma: ", i, gamma)

square = SplineGeometry()
square.AddRectangle([-0.6,-1],[0.6,1])
ngmesh = square.GenerateMesh(maxh=0.5**(i+1), quad_dominated=False)
mesh = Mesh (ngmesh)

# polynomial order in time
k_t = 2
# polynomial order in space
k_s = k_t
# spatial FESpace for solution
fes1 = H1(mesh, order=k_s)
# polynomial order in time for level set approximation
lset_order_time = k_t
# integration order in time
time_order = 2*k_t
# time finite element (nodal!)
tfe = ScalarTimeFE(k_t) 
# space-time finite element space
st_fes = SpaceTimeFESpace(fes1,tfe, flags = {"dgjumps": True})

#Fitted heat equation example
tend = 1
delta_t = tend/(2**(i+2))
tnew = 0

told = Parameter(0)
tref = ReferenceTimeVariable()
t = told + delta_t*tref

lset_adap_st = LevelSetMeshAdaptation_Spacetime(mesh, order_space = k_s, order_time = lset_order_time,
                                                threshold=0.05, discontinuous_qn=True)

# radius of disk (the geometry)
r0 = 0.5

# position shift of the geometry in time
rho =  CoefficientFunction((1/(pi))*sin(2*pi*t))
rhoL = lambda t:CoefficientFunction((1/(pi))*sin(2*pi*t))
# velocity of position shift
d_rho = CoefficientFunction(2*cos(2*pi*t))
#convection velocity:
w = CoefficientFunction((0,d_rho)) 

# level set
r = sqrt(x**2+(y-rho)**2)
levelset= r - r0

# solution and r.h.s.
Q = pi/r0   
u_exact = cos(Q*r) * sin(pi*t)
u_exactL = lambda t: cos(Q*sqrt(x**2+(y-rhoL(t))**2)) * sin(pi*t)
coeff_f = (Q/r * sin(Q*r) + (Q**2) * cos(Q*r)) * sin(pi*t) + pi * cos(Q*r) * cos(pi*t)


gfu = GridFunction(st_fes)

u_ic = CreateTimeRestrictedGF(gfu,0)
u_last = CreateTimeRestrictedGF(gfu,0)

u,v = st_fes.TnT()


lset_p1 = lset_adap_st.lset_p1    


lset_top = CreateTimeRestrictedGF(lset_p1,1.0)
lset_bottom = CreateTimeRestrictedGF(lset_p1,0.0)

dfm = lset_adap_st.deform
dfm_last_top = CreateTimeRestrictedGF(dfm,1.0)
dfm_current_top = CreateTimeRestrictedGF(dfm,1.0)
dfm_current_bottom = CreateTimeRestrictedGF(dfm,0.0)

t_old = 0

h = specialcf.mesh_size

Draw(lset_top,mesh,"lset")
Draw(IfPos(-lset_top,u_exactL(told),float('nan')),mesh,"u_exact")
Draw(IfPos(-lset_top,u_ic,float('nan')),mesh,"u")

Draw(dfm,mesh,"dfm")
Draw(dfm_current_bottom,mesh,"dfm_current_bottom")
Draw(dfm_current_top,mesh,"dfm_current_top")

visoptions.deformation = 1

lset_neg = { "levelset" : lset_p1, "domain_type" : NEG, "subdivlvl" : 0}
lset_neg_bottom = { "levelset" : lset_bottom, "domain_type" : NEG, "subdivlvl" : 0}
lset_neg_top = { "levelset" : lset_top, "domain_type" : NEG, "subdivlvl" : 0}

def SpaceTimeNegBFI(form):
    return SymbolicBFI(levelset_domain = lset_neg, form = form, time_order=time_order, definedonelements = ci.GetElementsOfType(HASNEG))

ci = CutInfo(mesh,time_order=time_order)

    
hasneg_integrators_a = []
hasneg_integrators_f = []
patch_integrators_a = []


hasneg_integrators_a.append(SpaceTimeNegBFI(form =  -u*(dt(v) + dt_vec(dfm)*grad(v))
                                            + delta_t*grad(u)*grad(v)
                                            - delta_t*u*InnerProduct(w,grad(v))))
hasneg_integrators_a.append(SymbolicBFI(levelset_domain = lset_neg_top, form = fix_t(u,1)*fix_t(v,1), deformation = dfm_current_top))

patch_integrators_a.append(SymbolicFacetPatchBFI(form = delta_t*(1+delta_t/h)*gamma*h**(-2)*(u-u.Other())*(v-v.Other()), skeleton=False, time_order=time_order))
hasneg_integrators_f.append(SymbolicLFI(levelset_domain = lset_neg, form = delta_t*coeff_f*v, time_order=time_order)) 
hasneg_integrators_f.append(SymbolicLFI(levelset_domain = lset_neg_bottom, form = u_ic*fix_t(v,0), deformation = dfm_current_bottom))

outfile = open("errorp"+str(k_t)+"_dev_i"+str(i)+"_gamma"+str(gamma)+".dat", "w")

l2max = 0

while tend - t_old > delta_t/2:
    # update lset geometry to new time slab (also changes lset_p1 !)
    #dfm = lset_adap_st.CalcDeformation(levelset,told,t_old,delta_t)
    with TaskManager():
        dfm = lset_adap_st.CalcDeformation(levelset,tref)
    
    RestrictGFInTime(spacetime_gf=lset_p1,reference_time=0.0,space_gf=lset_bottom)
    RestrictGFInTime(spacetime_gf=lset_p1,reference_time=1.0,space_gf=lset_top)
    RestrictGFInTime(spacetime_gf=dfm,reference_time=0.0,space_gf=dfm_current_bottom)   
    RestrictGFInTime(spacetime_gf=dfm,reference_time=1.0,space_gf=dfm_current_top)   

    if t_old == 0:
        mesh.SetDeformation(dfm_current_bottom)
        u_ic.Set(CoefficientFunction(u_exactL(0.)))
        mesh.UnsetDeformation()
    else:
        u_ic.Set(shifted_eval(u_last, back = dfm_last_top, forth = dfm_current_bottom))


    # update markers in (space-time) mesh
    ci.Update(lset_p1,time_order=time_order)

    # re-compute the facets for stabilization:
    ba_facets = GetFacetsWithNeighborTypes(mesh,a=ci.GetElementsOfType(HASNEG),
                                                b=ci.GetElementsOfType(IF))
    # re-evaluate the "active dofs" in the space time slab
    active_dofs = GetDofsOfElements(st_fes,ci.GetElementsOfType(HASNEG))

    # re-set definedonelements-markers according to new markings:
    for integrator in hasneg_integrators_a + hasneg_integrators_f:
        integrator.SetDefinedOnElements(ci.GetElementsOfType(HASNEG))
    for integrator in patch_integrators_a:
        integrator.SetDefinedOnElements(ba_facets)

    #a = BilinearForm(st_fes,check_unused=False,symmetric=False)
    a = RestrictedBilinearForm(st_fes,"a", ci.GetElementsOfType(HASNEG), ba_facets, check_unused=False)
    for integrator in hasneg_integrators_a + patch_integrators_a:
        a += integrator
    
    f = LinearForm(st_fes)
    for integrator in hasneg_integrators_f:
        f += integrator

    # update mesh deformation and assemble linear system
    mesh.SetDeformation(dfm)
    with TaskManager():
        a.Assemble()
        f.Assemble()
    mesh.UnsetDeformation()
    
    #mesh.SetDeformation(dfm_current_bottom)
    #f2.Assemble()
    #mesh.UnsetDeformation()    
    #f.vec.data = f.vec + f2.vec
    
    # solve linear system
    inv = a.mat.Inverse(active_dofs,inverse="pardiso")
    gfu.vec.data =  inv * f.vec

    # evaluate upper trace of solution for
    #  * for error evaluation 
    #  * upwind-coupling to next time slab
    RestrictGFInTime(spacetime_gf=gfu,reference_time=1.0,space_gf=u_last)   

    # update time variable (float and ParameterCL)
    t_old = t_old + delta_t
    told.Set(t_old)
    
    # compute error at final time
    mesh.SetDeformation(dfm_current_top)
    l2error = sqrt(Integrate(lset_neg_top,(u_exactL(told)-u_last)**2,mesh,order=2*k_s))
    mesh.UnsetDeformation()

    # print time and error
    #print("\rt = {0:10}, l2error = {1:20}".format(t_old,l2error),end="")
    print("t = ",t_old, " l2error = ",l2error)
    outfile.write(str(t_old)+"\t"+str(l2error)+"\n")
    
    if l2error > l2max:
        l2max = l2error
    
    # Redraw:
    Redraw(blocking=True)

    # store deformation at top level of for next time step
    dfm_last_top.vec.data = dfm_current_top.vec

print("L2 max: ", l2max)



"""
In this example we solve an *unfitted* Stokes interface problem with a
high order isoparametric unfitted discretisation method. This file is
based on the discretisations in cutfem.py, nxfem.py and
nxfem_higher_order.py.

Domain:
-------
The domain is [-1,1]^2 while the interface is described by a level set
function ( circle with radius R=2/3).

PDE problem:
------------
domain equations for velocity (u1, u2) and pressure p:
 mu_i div( sigma(u_i,p_i) ) = rho_i g  in sub-domain i, i=1,2,
                   div(u_i) = 0        in sub-domain i, i=1,2,
interface conditions:
                        [u] = 0        on interface (continuity across
                                                     the interface),
             [sigma(u,p)·n] = f        on interface (conservation of the
                                                     (momentum) flux),
                         u  = u_D      on domain boundary.

The r.h.s. term g corresponds to gravity, the term f is surface tension
force (here f = kappa · n where kappa is the mean curvature (1/R)). The
Dirichlet data is chosen according to match a manufactured solution
introduced in [1] which allows us to measure errors after the
computation of a discrete solution. The coefficients mu are domain-wise
constants which are different in the two sub-domains.

Discretisation:
---------------
* Finite element space: As in cutfem.py but for every velocity component
  and the pressure.

* Variational formulation: We use a Nitsche formulation which involves
  averages of the fluxes and jumps of the solution across the interface
  [2]. For the average we use the Hansbo-choice [3] where the average
  is adjusted to the local cut configuration in order to ensure stability
  of the resulting viscosity formulation. To ensure inf-sup for the
  velocity-pressure space we add a ghost penalty stabilization on the
  pressure space, cf. [2].

* Surface tension: In this example we prescribe the surface tension
  analytically, i.e. we circumvent approximating the mean curvature from
  the level set function.

Implementational aspects:
-------------------------
* Geometry approximation: As in nxfem_higher_order.py

* Ghost penalty stabilization: The edge-based stabilizations require
  different couplings (compared to standard discretisations). To add
  these to the sparsity pattern we have to add the "dgjumps" flags which
  prepares the sparse matrix for the corresponding needed couplings.

* Linear systems: A (sparse) direct solver is applied to solve the
  arising linear systems.

Literature:
-----------
[1] M. Kirchhart, S. Groß, A. Reusken, Analysis of an XFEM
    discretization for Stokes interface problems, SIAM J. Sci. Comput.,
    38(2), A1019–A1043, 2016.
[2] P. Lederer, C.-M. Pfeiler, C. Wintersteiger, C. Lehrenfeld, Higher
    order unfitted fem for Stokes interface problems. Proc. Appl. Math.
    Mech., 16: 7-10, 2016.
[3] A. Hansbo, P. Hansbo, An unfitted finite element method, based on
    Nitsche's method, for elliptic interface problems, Comp. Meth. Appl.
    Mech. Eng., 191(47):5537-5552, 2002
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.lsetcurv import *

SetHeapSize(100000000)

# -------------------------------- PARAMETERS ---------------------------------
# Mesh size
maxh = 0.125
# Finite element order (k of P^k/P^{k-1} Taylor-Hood elements)
order = 3

# Problem parameters
mu = [1, 10]

R = 2.0 / 3.0

aneg = 1.0 / mu[0]
apos = 1.0 / mu[1] + (1.0 / mu[0] - 1.0 / mu[1]) * exp(x * x + y * y - R * R)

# Discretisation parameters

# Stabilization parameters for ghost-penalty
gamma_stab_v = 0.05  # if set to zero: no GP-stabilization for velocity
gamma_stab_p = 0.05
# Stabilization parameter for Nitsche
lambda_nitsche = 0.5 * (mu[0] + mu[1]) * 20 * order * order

# ----------------------------------- MAIN ------------------------------------
# Generate the background mesh
square = SplineGeometry()
square.AddRectangle((-1, -1), (1, 1), bcs=[1, 2, 3, 4])
mesh = Mesh(square.GenerateMesh(maxh=maxh, quad_dominated=False))

d = mesh.dim


# Construct the exact level set function, radius and radius squared:
rsqr = x**2 + y**2
r = sqrt(rsqr)

levelset = r - R


# Construct the exact solution:
gammaf = 0.5  # surface tension = pressure jump
vel_exact = [a * CoefficientFunction((-y, x)) * exp(-1 * rsqr)
             for a in [aneg, apos]]
pre_exact = [x**3, x**3 - gammaf]


# Some helper expressions to compute the r.h.s. of the Stokes system for
# a given solution:
def Coef_Grad(v):
    func = tuple([v[i].Diff(w) for w in [x, y] for i in [0, 1]])
    return CoefficientFunction(func, dims=(d, d)).trans


def Coef_Eps(v):
    return 0.5 * (Coef_Grad(v) + Coef_Grad(v).trans)


def Coef_Div(v):
    func = tuple([sum([v[i, j].Diff(w) for j, w in [(0, x), (1, y)]])
                  for i in range(d)])
    return CoefficientFunction(func)


# Manufactured right-hand side:
coef_g = [Coef_Div(-2 * mu[i] * Coef_Eps(vel_exact[i])
                   + pre_exact[i] * Id(d)) for i in [0, 1]]


# Discretisation

# Higher order level set machinery
lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=10.5,
                                      discontinuous_qn=True)
deformation = lsetmeshadap.CalcDeformation(levelset)
lsetp1 = lsetmeshadap.lset_p1
ci = CutInfo(mesh, lsetp1)

# Finite element space
Vhbase = VectorH1(mesh, order=order, dirichlet=[1, 2, 3, 4])
Qhbase = H1(mesh, order=order - 1)

Vhneg = Compress(Vhbase, GetDofsOfElements(Vhbase,
                                           ci.GetElementsOfType(HASNEG)))
Vhpos = Compress(Vhbase, GetDofsOfElements(Vhbase,
                                           ci.GetElementsOfType(HASPOS)))
Qhneg = Compress(Qhbase, GetDofsOfElements(Qhbase,
                                           ci.GetElementsOfType(HASNEG)))
Qhpos = Compress(Qhbase, GetDofsOfElements(Qhbase,
                                           ci.GetElementsOfType(HASPOS)))

WhG = FESpace([Vhneg * Vhpos, Qhneg * Qhpos, NumberSpace(mesh)], dgjumps=True)

# Discrete functions
gfup = GridFunction(WhG)
gfu, gfp, gfn = gfup.components
gfu_neg, gfu_pos = gfu.components
gfp_neg, gfp_pos = gfp.components


# Element, facet and dof marking w.r.t. boundary approximation with lsetp1:
ba_facets = [GetFacetsWithNeighborTypes(mesh, a=ci.GetElementsOfType(HASNEG),
                                        b=ci.GetElementsOfType(IF)),
             GetFacetsWithNeighborTypes(mesh, a=ci.GetElementsOfType(HASPOS),
                                        b=ci.GetElementsOfType(IF))]

# Describe the integration regions
dx = tuple([dCut(lsetp1, dt, deformation=deformation) for dt in [NEG, POS]])
ds = dCut(lsetp1, IF, deformation=deformation)
dw = tuple([dFacetPatch(definedonelements=els_gp, deformation=deformation)
            for els_gp in ba_facets])


# Construct (Bi-)linear forms
h = specialcf.mesh_size
n_lset = 1.0 / Norm(grad(lsetp1)) * grad(lsetp1)
kappa = [CutRatioGF(ci), 1.0 - CutRatioGF(ci)]

a = BilinearForm(WhG, symmetric=False)
f = LinearForm(WhG)

u, p, n = WhG.TrialFunction()
v, q, m = WhG.TestFunction()


# Define some helper expressions:
def eps(u):
    return 0.5 * (Grad(u) + Grad(u).trans)


def sigma(i, u, p):
    return - 2 * mu[i] * eps(u[i]) + p[i] * Id(mesh.dim)


def average_flux(u, p):
    return sum([kappa[i] * sigma(i, u, p) * n_lset for i in range(2)])


def average_inv(u):
    return sum([kappa[1 - i] * u[i] for i in range(2)])


def jump(u):
    return u[0] - u[1]


# Stokes variational formulation:
for i in [0, 1]:  # Loop over domains
    a += 2 * mu[i] * InnerProduct(eps(u[i]), eps(v[i])) * dx[i]
    a += (- div(u[i]) * q[i] - div(v[i]) * p[i]) * dx[i]
    f += coef_g[i] * v[i] * dx[i]

# Constrain the pressure integral
a += (n * q[0] + m * p[0]) * dx[0]
# Nitsche parts:
a += (average_flux(u, p) * jump(v) + average_flux(v, p) * jump(u)) * ds
a += lambda_nitsche / h * jump(u) * jump(v) * ds
# Surface tension:
f += -gammaf * average_inv(v) * n_lset * ds

# Ghost-penalty terms:
for i in [0, 1]:
    if gamma_stab_v > 0:
        a += gamma_stab_v / h**2 * \
            (u[i] - u[i].Other()) * (v[i] - v[i].Other()) * dw[i]
    if gamma_stab_p > 0:
        a += -gamma_stab_p * \
            (p[i] - p[i].Other()) * (q[i] - q[i].Other()) * dw[i]


# Assemble and solve the resulting linear system
with TaskManager():
    a.Assemble()
    f.Assemble()

    # Set the in-homogeneous Dirichlet condition
    gfu.components[1].Set(vel_exact[1])

    # Solve linear system
    f.vec.data -= a.mat * gfup.vec

    gfup.vec.data += a.mat.Inverse(WhG.FreeDofs()) * f.vec

# Measure the error
err2 = {'vl2': 0, 'vh1': 0, 'pl2': 0.0}
for i in [0, 1]:
    _dx = dx[i].order(2 * order)
    err2['vl2'] += Integrate(Norm(gfu.components[i] - vel_exact[i])**2 * _dx,
                             mesh=mesh)
    err2['vh1'] += Integrate(Norm(Grad(gfu.components[i])
                                  - Coef_Grad(vel_exact[i]))**2 * _dx,
                             mesh=mesh)
    err2['pl2'] += Integrate(Norm(gfp.components[i] - pre_exact[i])**2 * _dx,
                             mesh=mesh)

print("L2 Error of velocity: {:10.8e}".format(sqrt(err2['vl2'])))
print("H1 Error of velocity: {:10.8e}".format(sqrt(err2['vh1'])))
print("L2 Error of pressure: {:10.8e}".format(sqrt(err2['pl2'])))

# Unset mesh adaptation
mesh.deformation = None

# Visualization of the solution
Draw(deformation, mesh, "deformation")
Draw(levelset, mesh, "levelset")
Draw(lsetp1, mesh, "lsetp1")
Draw(IfPos(lsetp1, gfu.components[1], gfu.components[0]), mesh, "vel")
Draw(IfPos(lsetp1, gfp.components[1], gfp.components[0]), mesh, "pressure")
